@@ -2,3 +2,5 @@
 
 #define MESSAGES_CGI "/cgi-bin/messages"
 #define DISPLAY_CGI "/cgi-bin/display"
+#define EDIT_CGI "/cgi-bin/edit"
+#define LISTADM_CGI "/cgi-bin/listadm"
