@@ -1,5 +1,5 @@
 /* Some useful functions...                                       */
-/* (c) 1997 by Bernhard Rosenkraenzer <root@startrek.in-trier.de> */
+/* (c) 1997 by Bernhard Rosenkraenzer <bero@bero-online.ml.org>   */
 
 #include "list.h"
 #include <stdio.h>
@@ -17,14 +17,22 @@ char *downcase(const char *s);
 char *strcasestr(const char *s1, const char *s2);
 char *strcasechr(const char *s, int c);
 char *salloc(size_t size);
+char *noleadingspaces(char *text);
 
 #ifdef WEIRD_STRCASECMP
-  int scasecmp(const char *s1, const char *s2); /* Fix a buggy implementation of strcasecmp */
+  int scasecmp(const char *s1, const char *s2);
+  int scmp(const char *s1, const char *s2);
 #else
   #define scasecmp(x,y) strcasecmp(x,y)
+  #define scmp(x,y) strcmp(x,y)
 #endif
 #ifdef WEIRD_STRLEN
-  size_t slen(const char *s);                   /* Fix a buggy implementation of strlen */
+  size_t slen(const char *s);
 #else
   #define slen(x) strlen(x)
 #endif  
+
+char exist(char *filename);
+char choice(char *prompt, char *choices);
+char yn(char *prompt);
+char *input(char *prompt, size_t maxsize);
